'
'Renard Vixen plug-in with chipiplexing extensions
'
'Base Renard version: Copyright (c) 2005-2009, K.C. Oaks
'Chipiplexing extensions: Copyright (c) 2009 - 2010 Don Julien.  djulien@thejuliens.net
'I can't speak for the base version itself, but the chipiplexing extensions may be freely 
'distributed and used for non-commerical purposes only.  See the license agreement terms below.
' 
'Revision history:
'1.0  12/15/09  DJ  first stable version used in the show
'1.0a 12/18/09  DJ  add PauseOnError option; fixed array subscript
'1.0b 12/29/09  DJ  add PIC pin swapping (temp kludge; board #2 pinout does not match board #1, d'oh)
'1.0c 1/9/10    DJ  misc fixes for min = 0 and max = 255 (for visible LEDs)
'1.0d 1/26/10   DJ  make compatible with multiple versions of Vixen (2.1.1.0 and 2.5.0.8); misc cleanup
'
' SOFTWARE LICENSE AGREEMENT:
' This extended software is licensed under the Creative Commons Attribution-Noncommercial-Share Alike 
' 3.0 Unported License.  To view a copy of this license, visit http://creativecommons.org/licenses/by-nc-sa/3.0/ 
' or send a letter to Creative Commons, 171 Second Street, Suite 300, San Francisco, California, 94105, USA.
'
' Basically, you may use this software freely for your own personal, non-commerical use, as long as you also share
' freely any changes, adaptations, or improvements you make.  Of course, the standard conditions also apply:
' THIS SOFTWARE IS PROVIDED "AS IS".  NO EXPRESS, IMPLIED, STATUTORY OR OTHER WARRANTIES 
' OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE APPLY TO THIS SOFTWARE.  AUTHOR 
' SHALL NOT BE LIABLE FOR SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES, FOR ANY REASON OR
' UNDER ANY CIRCUMSTANCES WHATSOEVER.  THIS SOFTWARE MUST NOT BE USED FOR CRITICAL APPLICATIONS
' SUCH AS HEAVY MACHINERY, LIFE SUPPORT, ETC.  YOU USE THIS SOFTWARE AT YOUR OWN RISK.
' But if you have problems or questions, let me know and I'll try to help if I can.


'
' this thing started with the Renard plugin packaged with Vixen 2.1.1.0,
' then extended to support chipiplexing,
' then merged in changes from the Renard plug packaged with Vixen 2.5.0.8
'
'for some sketchy info about Vixen plugins, see:
'http://www.doityourselfchristmas.com/forums/showthread.php?t=3929
'http://www.doityourselfchristmas.com/forums/showthread.php?t=50324
'http://www.doityourselfchristmas.com/forums/showthread.php?t=722
'existing plug-ins can be disassembled by using .NET Reflector from 
'http://www.red-gate.com/products/reflector/
'this is a VERY useful tool!


Imports System
'Imports System.Collections.Generic
Imports System.IO.Ports
'Imports System.IO.File 
Imports System.Text
'Imports System.Runtime.InteropServices
Imports System.Windows.Forms
Imports System.Threading
Imports System.Xml
Imports System.Reflection


Imports Vixen 'this will be 2.1 or 2.5, depending on the project references

'dummy interface to force Vixen to load this DLL:
'Public Interface IOutputPlugIn
'	Sub DummyFunction()
'End Interface


'main implementation class:
Public Class RenardChipiplexerPlugin
	Implements Vixen.IEventDrivenOutputPlugIn

#Region "Internal data"
'generic Renard data:
	Private m_channelValues As Byte()
	Private m_dialogList As List(Of Form) = New List(Of Form)
	Private m_eventThread As Thread
	Private m_eventTrigger As AutoResetEvent
	Private m_holdPort As Boolean
	Private m_p1Packet As Byte() = New Byte(18  - 1) {}
	Private m_p1Zeroes As Byte()
	Private m_p2Packet As Byte()
	Private m_p2Zeroes As Byte()
	Private m_protocolVersion As Integer = 1
	Private m_running As Boolean = False
	Private WithEvents m_selectedPort As SerialPort = Nothing
	Private m_setupData As Vixen.SetupData
	Private m_setupNode As XmlNode
    Private m_threadPosition As Integer = -1
	Private Const PAD_DISTANCE As Integer = 100
	Private Enum RunState '2.5.0.8
		Running = 0
		Stopped = 2
		Stopping = 1
	End Enum
	Private m_state As RunState '2.5.0.8
'added a few more generic data items: -dj
	Private m_logfile As String = String.Empty 'added for easier debug/performance analysis
	Private m_errpause As Boolean 'stop on error, or just log it and keep going
	Private m_trace As Boolean 'log all data (mainly for debug or perf tuning)
	Private ReadOnly LogFileContention As New Object 'serialize writes to log file to avoid threading problems
	Private m_starttime As Date 'log entry times are relative to start of execution
#If WANT_CHIPIPLEXING Then 'turn this OFF if you don't want chipiplexing
'chipiplexing plug-in options:
	Private CfgUnused As Integer = &h10 'an unused config bit
	Private m_version As Integer = 1 'in case logic changes, we want backwards-compatibility
	Private m_minbright As Byte = 20 'treat anything below this as full off
	Private m_maxbright As Byte = 205 'treat anything above this as full on
	Private m_fullbright As Byte = 204 'above this will cause flicker (too soon in AC cycle for triac to latch)
	Private m_maxrpl As Integer = 2 'max #rows per brightness level
	Private m_closeness As Integer = 3 'treat brightness levels within this difference as the same, if needed in order to meet chipiplexing restrictions (max data size or #rows per level)
	Private m_nodiscard As Boolean = True 'do not discard any data; throws an error instead
	Private m_bufdedup As Integer = -1 'don't send out duplicate buffers to individual PICs
	Private m_renbytes As Boolean = False 'avoid Renard bytes in output
	Private m_maxoutlen As Integer = 80 'max output buffer length
#If BOARD1_INCOMPAT Then
	Private m_options As String = "1.0d BOARD1 INCOMPAT"
#Else
	Private m_options As String = "1.0d"
#End If
'chipiplexing data:
	Private m_inbuf(100 - 1) As Byte 'not required for chipiplexing, but useful for comm problem diagnosis
	Private m_outbuf(2+2+200 - 1) As Byte 'controller output buffer; TODO: config 56 or 80
	Private m_prevbuf(2+2+200 - 1) As Byte 'previous output buffer
	Private m_rcvbuf As String = String.Empty 'debug/perf data coming back from PICs
	Private m_prevbuflen As Integer 'length of previous output buffer
	Private m_numch As Integer = 0 '#channels to process
'this would be the preferable representation:
'	Structure LevelInfo
''		Dim level As Byte 'brightness levels of each channel
'		Dim rowindex As Byte 'rows using this brightness level
'		Dim columns(8 - 1) As Byte 'columns in each row
''		Dim numrows As Byte '#rows for this brightness level
'	End Structure
'	Private m_levels(56 - 1) as LevelInfo 
'VB.NET doesn't allow access to some parts of this struct, so just use separate arrays:
	Private m_rowindex(56 - 0) As Byte 'rows using this brightness level
	private m_columns(56 - 0, 8 - 1) As Byte 'columns in each row
	Private m_brindex(256 - 1) As Byte 'map brightness level to an entry in levels/rowindex/columns
	Private num_levels As Integer '#brightness levels in this event
	Private NumBits(256 - 1) As Byte 'lookup table for #bits in a byte; this is faster than computing it
'chipiplexing firmware options:
	Private m_cfg As Byte = &h02 '2x pulse width (16 usec)
'performance stats:
	Private total_evts As Long 'total #events processed
	Private total_inlen As Long 'total data length received for processing
	Private total_outlen As Long 'total data length sent
	Private protocol_outlen As Long 'output length due to protocol
	Private total_realnulls As Long 'total #channel bytes that were full off (not sent)
	Private total_nearnulls As Long 'total #channel bytes that were treated as full off (not sent)
	Private total_nearfulls As Long 'total #channel bytes that were treated as full on
	Private total_levels As Long 'total #brightness levels sent
	Private total_rowsoverflowed As Long '#rows overflowed into a close brightness level (lossy dimming)
	Private total_levelsmerged As Long '#levels merged with a close brightness level (lossy dimming)
	Private total_rowsdropped As Long '#rows discarded (lossy dimming)
	Private total_levelsdropped As Long '#levels discarded (lossy dimming)
	Private total_rows As Long 'total #rows processed
	Private total_columns As long 'total #columns processed
	Private max_outlen As Integer 'max data length sent in a frame
	Private max_levels As Integer 'max #levels in a frame
	Private max_rows As Integer 'max #rows in a frame
	Private max_columns As Integer 'max #columns in a frame
	Private min_outlen As Integer 'min data length sent in a frame
	Private min_levels As Integer 'min #levels in a frame
	Private min_rows As Integer 'min #rows in a frame
	Private min_columns As Integer 'min #columns in a frame
	Private num_dups As Integer '#duplicate buffers (not sent)
	Private num_avoids As Integer '#special protocol bytes avoided
#End If
'special chars in Renard protocol:
	Private Const Renard_SafeByte As Byte = &h7C
	Private Const Renard_PadByte As Byte = &h7D
	Private Const Renard_SyncByte As Byte = &h7E
	Private Const Renard_EscByte As Byte = &h7F
	Private Const Renard_CmdAdrsByte As Byte = &h80
'	Private Renard_Reserved As Byte() = New Byte() {Renard_SafeByte, Renard_PadByte, Renard_SyncByte, Renard_EscByte, Renard_CmdAdrsByte}
'	Private Renard_Reserved As Byte() = New Byte() {Renard_SyncByte} 'this is the only one that matters if regular Renards are not present
'	Private RenardReserved_lowest As Byte 'lowest reserved byte value (assumed to be consecutive)
'	Private RenardReserved_highest as Byte 'highest reserved byte value (assumed to be consecutive)
#End Region

#Region "Vixen version-dependent stuff"
'KLUDGE: Vixen 2.5.0.8 is incompatible with 2.1.1.0; IOutputPlugin.Setup() changed return types.
'standard interface compatibility conventions are badly broken in Vixen,
'so we need an ugly work around to fix this problem.

#If VIXEN = 21
	Private Const m_vixver As String = "2.1.x"
'interface broken (return type different between 2.1 and 2.5)
	Function StartUp_21 As list (of system.Windows.Forms.form) Implements Vixen.IEventDrivenOutputPlugIn.StartUp
		Me.StartUp_any
        Return Me.m_dialogList
	End Function
'interface strange (this one needs to be implemented 2x)
	Overrides Function ToString() As String Implements Vixen.IOutputPlugIn.ToString
		Return Me.Name_any
	End Function
#End If
#If VIXEN = 25
	Private Const m_vixver As String = "2.5.x"
'interface broken (return type different between 2.1 and 2.5)
	Sub StartUp_25 Implements Vixen.IEventDrivenOutputPlugIn.StartUp
		Me.StartUp_any
	End Sub
#End If
#End Region

#Region "Instantiation"
	Public Sub New()
'TODO: decide whether it is better to statically allocate memory in declares, or dynamically here (it probably doesn't matter)
		On Error GoTo errh 'an exception in here will cause Vixen to say it couldn't find the plug-in; add a little error handling to make that less confusing
		Me.m_selectedPort = Nothing '2.5.0.8
		Me.m_protocolVersion = 1 '2.5.0.8
'pre-fill/pre-allocate protocol 1 + 2 buffers:
		Me.m_p1Packet = New Byte(18  - 1) {} '2.5.0.8
		Me.m_p1Packet(0) = Renard_SyncByte
		Me.m_p2Packet = New Byte(11  - 1) {}
		Me.m_p2Packet(0) = 0
		Me.m_p1Zeroes = New Byte(16  - 1) {}
		Me.m_p2Zeroes = New Byte(8  - 1) {}
'thread setup:
		Me.m_state = RunState.Stopped '2.5.0.8
'		Me.m_eventThread = New Thread(New ThreadStart(AddressOf Me.EventThread)) '2.1.1.0
'		Me.m_eventTrigger = New AutoResetEvent(False) '2.1.1.0
'misc setup:
		Dim i As Integer, bit As Integer
'		Me.RenardReserved_lowest = 255
'		Me.RenardReserved_highest = 0
'		For i = 1 To Me.Renard_Reserved.Length
'			If Me.Renard_Reserved(i-1) < Me.RenardReserved_lowest Then Me.RenardReserved_lowest = Me.Renard_Reserved(i-1)
'			If Me.Renard_Reserved(i-1) > Me.RenardReserved_highest Then Me.RenardReserved_highest = Me.Renard_Reserved(i-1)
'		Next i
'compute #bits once at the start rather than maintaining an ugly hard-coded lookup table:
		For i = 0 To 255
			Me.NumBits(i) = 0
			For bit = 0 To 7
				If i And (1<<bit) Then Me.NumBits(i) += 1
			Next
		Next i

		Exit Sub
errh:
		ReportError("ctor")
	End Sub
#End Region

#Region "Properties"
	Private ReadOnly Property Name_any As String Implements Vixen.IEventDrivenOutputPlugIn.Name
		Get
'			Dim vixen As Reflection.Assembly = Reflection.Assembly.GetEntryAssembly 'GetCallingAssembly=mscorlib
'			Select Case vixen.GetName.Version.ToString
			Return "Renard (chipiplexed) " & m_vixver
		End Get
	End Property

	Private ReadOnly Property Description_any As String Implements Vixen.IEventDrivenOutputPlugIn.Description
		Get
			Return "Phil Short's Renard Dimmer output plugin, with chipiplexing extensions"
		End Get
	End Property

	Private ReadOnly Property Author_any As String Implements Vixen.IEventDrivenOutputPlugIn.Author
		Get
			Return "K.C. Oaks and Phil Short (base code), DIYC::djulien (chipiplexing extensions)"
		End Get
	End Property

'"The plug-in is expected to provide an array of which PortType resources it is using by index. Currently
'only supports four serial And three parallel ports."
	Private ReadOnly Property HardwareMap_any As Vixen.HardwareMap() Implements Vixen.IEventDrivenOutputPlugIn.HardwareMap
		Get
        	On Error Goto errh 'added error handling -dj
'fixed up port# parsing -dj
'            Return New HardwareMap() { New HardwareMap("Serial", Integer.Parse(Me.m_selectedPort.PortName.Substring(3))) }
			Return New Vixen.HardwareMap() { New Vixen.HardwareMap("Serial", "0" & Integer.Parse(Me.m_selectedPort.PortName.Replace("COM", ""))) }
errh:
			ReportError("HWMap")
		End Get
	End Property

'"Vixen expects that a plug-in will provide a ToString override returning the Name member. The return value
'of ToString will be what's displayed in any listboxes in which the plug-In Is listed."
	Private Function ToString_any() As String Implements Vixen.IEventDrivenOutputPlugin.ToString', Vixen.IOutputPlugIn.ToString
'		Return "Renard (chipiplexed) plugin for Vixen " & m_vixver
		Return Me.Name_any
	End Function
#End Region

#Region "Vixen event handlers"
'"Called anytime Vixen needs to make sure the plug-in's setup or other initialization is up to date.
'Initialize is called before the plug-in is setup, before sequence execution, and other times. It's called
'from multiple places at any time, therefore the plug-in can make no assumptions about the state of the
'program or sequence due to a call to Initialize.
'	channels: An array of Channel objects representing the channels present in the owning sequence.
'	setupData: A SetupData reference that provides some plug-incentric convenience methods. It can be safely ignored.
'	setupNode: An XmlNode representing the root of the plug-in's setup data in the sequence. Please see the Vixen.Xml class for some convenience methods."
	Private Sub Initialize_any(ByVal executableObject As Vixen.IExecutable, ByVal setupData As Vixen.SetupData, ByVal setupNode As XmlNode) Implements Vixen.IEventDrivenOutputPlugIn.Initialize
		On Error Goto errh 'added error handling -dj
		Me.m_setupData = setupData
		Me.m_setupNode = setupNode
'expanded this line for easier debug: -dj
'		Me.m_selectedPort = New SerialPort(Me.m_setupData.GetString(Me.m_setupNode, "name", "COM1"), Me.m_setupData.GetInteger(Me.m_setupNode, "baud", 19200), DirectCast(Enum.Parse(GetType(Parity), Me.m_setupData.GetString(Me.m_setupNode, "parity", Parity.None.ToString)), Parity), Me.m_setupData.GetInteger(Me.m_setupNode, "data", 8), DirectCast(Enum.Parse(GetType(StopBits), Me.m_setupData.GetString(Me.m_setupNode, "stop", StopBits.One.ToString)), StopBits))
		Dim portName As String = Me.m_setupData.GetString(Me.m_setupNode, "name", "COM1")
		Dim baudRate As Integer = Me.m_setupData.GetInteger(Me.m_setupNode, "baud", 19200)
		Dim parityy As Parity = DirectCast(System.Enum.Parse(GetType(Parity), Me.m_setupData.GetString(Me.m_setupNode, "parity", Parity.None.ToString)), Parity)
		Dim dataBits As Integer = Me.m_setupData.GetInteger(Me.m_setupNode, "data", 8)
		Dim stopBitss As StopBits = DirectCast(System.Enum.Parse(GetType(StopBits), Me.m_setupData.GetString(Me.m_setupNode, "stop", StopBits.One.ToString)), StopBits)
		Me.m_selectedPort = New SerialPort(portName, baudRate, parityy, dataBits, stopBitss)
		If Me.m_selectedPort Is Nothing Then ReportError("Init: Unable to get serial port '" & portName & "', baud " & baudRate)
		Me.m_protocolVersion = Me.m_setupData.GetInteger(Me.m_setupNode, "ProtocolVersion", 1)
		Me.m_holdPort = Me.m_setupData.GetBoolean(Me.m_setupNode, "HoldPort", True)
		Me.m_selectedPort.WriteTimeout = 500
		Me.m_logfile = Me.m_setupData.GetString(Me.m_setupNode, "LogFile", String.Empty)
		Me.m_errpause = Me.m_setupData.GetBoolean(Me.m_setupNode, "PauseOnError", False)
		Me.m_trace = Me.m_setupData.GetBoolean(Me.m_setupNode, "DataLogging", False)
		Me.m_version = Me.m_setupData.GetString(Me.m_setupNode, "PluginVersion", 1)
#If WANT_CHIPIPLEXING Then
		Me.m_numch = Integer.Parse("0" & m_setupnode.Attributes("to").InnerText) - Integer.Parse("0" & m_setupnode.Attributes("from").InnerText) + 1
'		If (Me.m_numch Mod 56) Or (Me.m_numch < 1) Then msgbox("Must be a multiple of 56 channels (" & Me.m_numch & " found); this is a temporary limitation in this version.", MsgBoxStyle.Exclamation, "CAUTION/TODO")
		If Me.m_protocolVersion = 3 Then 'chipiplexing
'get extra chipiplexing config info:
			Me.m_cfg = Me.m_setupData.GetInteger(Me.m_setupNode, "ChipiplexingConfigByte", 1)
			Me.m_minbright = Me.m_setupData.GetInteger(Me.m_setupNode, "ChipiplexingMinBright", 40)
			Me.m_maxbright = Me.m_setupData.GetInteger(Me.m_setupNode, "ChipiplexingMaxBright", 200)
			Me.m_fullbright = Me.m_setupData.GetInteger(Me.m_setupNode, "ChipiplexingFullBright", 204)
			Me.m_maxrpl = Me.m_setupData.GetInteger(Me.m_setupNode, "ChipiplexingMaxRowsPerLevel", 2)
			Me.m_closeness = Me.m_setupData.GetInteger(Me.m_setupNode, "ChipiplexingCloseness", 3)
			Me.m_nodiscard = Not Me.m_setupData.GetBoolean(Me.m_setupNode, "ChipiplexingLossy", False)
			Me.m_bufdedup = Me.m_setupData.GetInteger(Me.m_setupNode, "ChipiplexingDeDup", -1)
			Me.m_renbytes = Me.m_setupData.GetBoolean(Me.m_setupNode, "RenardBytes", False)
			Me.m_maxoutlen = Me.m_setupData.GetInteger(Me.m_setupNode, "MaxOutlen", 80)
'initialize chipiplexing data/stats:
			Me.total_evts = 0 'total #events processed
			Me.total_inlen = 0 'total data length received for processing
			Me.total_outlen = 0 'total data length sent
			Me.protocol_outlen = 0 'output length due to protocol
			Me.total_realnulls = 0 'total #channel bytes that were full off (not sent)
			Me.total_nearnulls = 0 'total #channel bytes that were treated as full off (not sent)
			Me.total_nearfulls = 0 'total #channel bytes that were treated as full on
			Me.total_rowsoverflowed = 0 '#rows overflowed into a close brightness level (lossy dimming)
			Me.total_levelsmerged = 0 '#levels merged with a close brightness level (lossy dimming)
			Me.total_rowsdropped = 0 '#rows discarded (lossy dimming)
			Me.total_levelsdropped = 0 '#levels discarded (lossy dimming)
			Me.total_rows = 0 'total #rows processed
			Me.total_columns = 0 'total #columns processed
			Me.max_outlen = 0 'max data length sent in a frame
			Me.max_levels = 0 'max #levels in a frame
			Me.max_rows = 0 'max #rows in a frame
			Me.max_columns = 0 'max #columns in a frame
			Me.min_outlen = Me.m_outbuf.Length + 1 'min data length sent in a frame
			Me.min_levels = 256 'min #levels in a frame
			Me.min_rows = 256 'min #rows in a frame
			Me.min_columns = 256 'min #columns in a frame
			Me.num_dups = 0 '#duplicate buffers (not sent)
			Me.num_avoids = 0 '#special protocol bytes avoided
			Me.m_prevbuflen = 0 'length of previous output buffer
		End If
#End If
		Exit Sub
errh:
		ReportError("Init")
	End Sub

'"Called when the user has requested to setup the plug-in instance."
'entry in xml file looks something like this:
'<PlugIn name="Renard Dimmer (modified)" key="935429715" id="0" enabled="False" from="1" to="296">
	Private Sub Setup_any() Implements Vixen.IEventDrivenOutputPlugIn.Setup
		On Error Goto errh 'added error handling -dj
		Me.m_numch = Integer.Parse("0" & m_setupnode.Attributes("to").InnerText) - Integer.Parse("0" & m_setupnode.Attributes("from").InnerText) + 1
		Dim dialog As RenardChipiplexerSetup = New RenardChipiplexerSetup(Me.m_selectedPort, Me.m_protocolVersion, Me.m_holdPort, Me.m_logfile, Me.m_errpause, Me.m_trace, Me.m_cfg, Me.m_fullbright, Me.m_maxbright, Me.m_minbright, Me.m_maxrpl, Me.m_closeness, Not Me.m_nodiscard, Me.m_bufdedup, Me.m_renbytes, Me.m_maxoutlen, Me.m_rcvbuf)
		Dim result As Integer = dialog.ShowDialog
		If (result = DialogResult.OK) Then
			Me.m_selectedPort = dialog.SelectedPort
			Me.m_setupData.SetString(Me.m_setupNode, "name", Me.m_selectedPort.PortName)
			Me.m_setupData.SetInteger(Me.m_setupNode, "baud", Me.m_selectedPort.BaudRate)
			Me.m_setupData.SetString(Me.m_setupNode, "parity", Me.m_selectedPort.Parity.ToString)
			Me.m_setupData.SetInteger(Me.m_setupNode, "data", Me.m_selectedPort.DataBits)
			Me.m_setupData.SetString(Me.m_setupNode, "stop", Me.m_selectedPort.StopBits.ToString)
			Me.m_protocolVersion = dialog.ProtocolVersion
			Me.m_setupData.SetInteger(Me.m_setupNode, "ProtocolVersion", Me.m_protocolVersion)
			Me.m_holdPort = dialog.HoldPort
			Me.m_setupData.SetBoolean(Me.m_setupNode, "HoldPort", Me.m_holdPort)
			Me.m_logfile = dialog.LogFile
			Me.m_setupData.SetString(Me.m_setupNode, "LogFile", Me.m_logfile)
			Me.m_errpause = dialog.PauseOnError 
			Me.m_setupData.SetBoolean(Me.m_setupNode, "PauseOnError", Me.m_errpause)
			Me.m_trace = dialog.DataLogging 
			Me.m_setupData.SetBoolean(Me.m_setupNode, "DataLogging", Me.m_trace)
			Me.m_setupData.SetInteger(Me.m_setupNode, "PlugInVersion", Me.m_version)
			If Me.m_protocolVersion = 3 Then
				If ((Me.m_numch Mod 56) <> 0) Or (Me.m_numch < 1) Then MsgBox(String.Format("Number of channels ({0}) is not a multiple of 56.  SSR channels may not be distributed how you expect on the last PIC.", Me.m_numch), MsgBoxStyle.Exclamation, "CAUTION")

				Me.m_cfg = dialog.ChipiplexingConfig
				If Me.IsRenardByte(Me.m_cfg) Then 
					Me.m_cfg = Me.m_cfg XOr Me.CfgUnused 'KLUDGE: change an unused config bit to avoid special protocol chars
					If Me.IsRenardByte(Me.m_cfg) Then throw new Exception(String.Format("Using alternate config byte {0} still creates protocol byte conflict. Please report this.", Me.m_cfg))
					MsgBox(String.Format("Config byte was changed to {0} to avoid conflict with a protocol byte.", Me.m_cfg), MsgBoxStyle.Exclamation, "WARNING")
				end if
				Me.ParanoidSetInteger("ChipiplexingConfigByte", Me.m_cfg)

				me.m_fullbright = dialog.ChipiplexingFullBright 
				Me.ParanoidSetInteger("ChipiplexingFullBright", Me.m_fullbright)

				me.m_maxbright = dialog.ChipiplexingMaxBright 
				Me.ParanoidSetInteger("ChipiplexingMaxBright", Me.m_maxbright)

				me.m_minbright = dialog.ChipiplexingMinBright 
				Me.ParanoidSetInteger("ChipiplexingMinBright", Me.m_minbright)

				Me.m_maxrpl = dialog.ChipiplexingMaxRowsPerLevel
				Me.ParanoidSetInteger("ChipiplexingMaxRowsPerLevel", Me.m_maxrpl)

				Me.m_closeness = dialog.ChipiplexingCloseness
				Me.ParanoidSetInteger("ChipiplexingCloseness", Me.m_closeness)

				Me.m_nodiscard = Not dialog.ChipiplexingLossy
				Me.m_setupData.SetBoolean(Me.m_setupNode, "ChipiplexingLossy", Not Me.m_nodiscard)

				Me.m_bufdedup = dialog.ChipiplexingDedup 
				Me.m_setupData.SetInteger(Me.m_setupNode, "ChipiplexingDeDup", Me.m_bufdedup)

				Me.m_renbytes = dialog.ChipiplexingRenBytes
				Me.m_setupData.SetBoolean(Me.m_setupNode, "RenardBytes", Me.m_renbytes)
				
				Me.m_maxoutlen = dialog.ChipiplexingMaxOutlen
				Me.ParanoidSetInteger("MaxOutlen", Me.m_maxoutlen)
'			Else
'				MsgBox(Me.m_protocolVersion & " vs. " & 2)
			End If
		Else 'explicitly warn user that data was not saved
			MsgBox(String.Format("NOT saved ({0}:{1} vs. {2}:{3}).", TypeName(result), result, TypeName(DialogResult.OK), DialogResult.OK), MsgBoxStyle.Information, "NOT Saved")
		End If
		dialog.Dispose
		Exit Sub
errh:
		ReportError("Setup")
	End Sub

'"Called when a sequence is executed. Returns a list of forms that the plug-in needs to have injected
'into Vixen's MDI interface. The previews are examples Of plug-ins that Return forms For this purpose."
'Vixen interface is broken (function prototype is different between Vixen 2.1 and 2.5), so a separate version-specific wrapper function will call this generic function
	Private Sub Startup_any() 'As List(Of Form) 'Implements Vixen2508.IHardwarePlugin.Startup, Vixen.IOutputPlugIn.Startup '2.1.1.0
		On Error Goto errh 'added error handling -dj
        If Me.m_selectedPort Is Nothing Then Return 'Me.m_dialogList 'nothing to do
        Me.m_starttime = DateTime.Now 
		Dim ver As Version = Reflection.Assembly.GetEntryAssembly.GetName.Version 'GetCallingAssembly=mscorlib 'GetExecutingAssmebly=this dll
		If Me.m_trace Then LogMsg(String.Format("========== Startup at {5:MM-dd-yyyy HH:mm:ss}.{6:D3}, Vixen {0}.{1}.{2}.{3}, plug-in {4} ==========", New Object() { ver.Major, ver.Minor, ver.MajorRevision, ver.MinorRevision, Me.m_options, DateTime.Now, DateTime.Now.TimeOfDay.Milliseconds}))
        If Not (Not Me.m_holdPort OrElse Me.m_selectedPort.IsOpen) Then Me.m_selectedPort.Open
		Me.m_selectedPort.Handshake = Handshake.None
		Me.m_selectedPort.Encoding = Encoding.UTF8
		Me.m_selectedPort.RtsEnable = True
		Me.m_selectedPort.DtrEnable = True
		If Me.m_holdPort Then
'			Me.m_running = True '2.1.1.0
'			Me.m_eventThread.Start '2.1.1.0
			Dim th As Thread = New Thread(New ThreadStart(AddressOf Me.EventThread))
			th.Start()
			Do While (Me.State <> RunState.Running) '2.5.0.8
				Thread.Sleep(1) '2.5.0.8
			Loop '2.5.0.8
		End If
		Return 'Me.m_dialogList 
errh:
		ReportError("Startup")
	End Sub


'"Called when the plug-in needs to update the hardware state during execution.
'	channelValues: Event values in channel order, 1 byte per channel."
'http://www.doityourselfchristmas.com/forums/showthread.php?p=37947
'"The event method is called whenever there is a change in the state of the channel values. With each call, the method receives
'the data For every channel that's being sent so that plugin."
	Private Sub Event_any(ByVal channelValues As Byte()) Implements Vixen.IEventDrivenOutputPlugIn.[Event]
		On Error Goto errh 'added error handling -dj
        If Me.m_selectedPort Is Nothing Then Exit Sub 'nothing to do
'CAUTION: event may be processed asynchronously, so channel bytes need to be copied to local memory; extra memory copying overhead
		Me.m_channelValues = channelValues 
		If Me.m_holdPort Then
			Me.m_eventTrigger.Set
		Else
			If Not Me.m_selectedPort.IsOpen Then Me.m_selectedPort.Open
			Me.FireEvent
			Me.m_selectedPort.Close
		End If
		Exit Sub
errh:
		ReportError("Event")
	End Sub

'"Called when execution is stopped or the plug-in instance is no longer going To be referenced."
	Private Sub Shutdown_any() Implements Vixen.IEventDrivenOutputPlugIn.Shutdown
		On Error Goto errh 'added error handling -dj
		If Me.m_trace Then LogMsg(String.Format("========== Shutdown at {0:MM-dd-yyyy HH:mm:ss}.{1:D3} ==========", New Object() { DateTime.Now, DateTime.Now.TimeOfDay.Milliseconds}))
#If WANT_CHIPIPLEXING Then
		If (Me.m_protocolVersion = 3) And Me.m_trace Then 'report chipiplexing data/stats
			LogMsg("Chipiplexing config:")
			LogMsg(String.Format("Channels {0}, min bright {1}, max bright {2}, full bright {3}, max rows/level {4}, close {5}, discard {6}, dedup {7}, renard {8}, max len {9}", Me.m_numch, Me.m_minbright, Me.m_maxbright, Me.m_fullbright, Me.m_maxrpl, Me.m_closeness, Me.m_nodiscard, Me.m_bufdedup, Me.m_renbytes, Me.m_maxoutlen))
			LogMsg("Chipiplexing stats:")
			LogMsg(String.Format("Total events {0}, total bytes in {1}, total bytes out {2}, for protocol {3} ({4}% of inlen), for data {5} ({6}% of inlen), min data {7}, max data {8}, average data {9}, dup bufs(=) {10}, protocol avoids(@) {11}", Me.total_evts, Me.total_inlen, Me.total_outlen, Me.protocol_outlen, SafeDiv(100*Me.protocol_outlen, Me.total_inlen), Me.total_outlen - Me.protocol_outlen, SafeDiv(100*(Me.total_outlen - Me.protocol_outlen), Me.total_inlen), Me.min_outlen, Me.max_outlen, SafeDiv((Me.total_outlen - Me.protocol_outlen), Me.total_evts), Me.num_dups, Me.num_avoids))
			LogMsg(String.Format("Off channels (not sent) {0}, near-off channels (not sent) {1}, near-full channels (treated as full on) {2}", Me.total_realnulls, Me.total_nearnulls, Me.total_nearfulls))
			LogMsg(String.Format("Min levels/rows/columns per event {0}, {1}, {2}, max {3}, {4}, {5}, average {6}, {7}, {8}", Me.min_levels, Me.min_rows, Me.min_columns, Me.max_levels, Me.max_rows, Me.max_columns, SafeDiv(Me.total_levels, Me.total_evts), SafeDiv(Me.total_rows, Me.total_evts), SafeDiv(Me.total_columns, Me.total_evts)))
			LogMsg(String.Format("Levels merged(^) {0}, rows overflowed(*) {1}, levels discarded(!) {2}, rows discarded(-) {3}", Me.total_levelsmerged, Me.total_rowsoverflowed, Me.total_levelsdropped, Me.total_rowsdropped))
		End If
#End If
'		If Me.m_running Then '2.1.1.0
'			Me.m_running = False '2.1.1.0
		If (Me.State = RunState.Running) Then
			Me.State = RunState.Stopping
'			If Not Me.m_eventTrigger Is Nothing Then '2.1.1.0
'				Me.m_eventTrigger.Set
'				Do While (Me.m_threadPosition <> 4)
'					Application.DoEvents
'				Loop
'				If Not Me.m_eventThread.Join(1000) Then
'					MessageBox.Show("Event thread did not terminate in time!")
'				End If
'			End If
			Do While (Me.State <> RunState.Stopped) '2.5.0.8
				Thread.Sleep(5) '2.5.0.8
			Loop '2.5.0.8
		End If
'		Try 
			If Not Me.m_selectedPort Is Nothing Then 
				If Me.m_selectedPort.IsOpen Then Me.m_selectedPort.Close
			End If
'		Catch exception As Exception
'			MessageBox.Show(("Port close: " & exception.Message))
'		End Try
		Exit Sub
errh:
		ReportError("Shutdown")
	End Sub
#End Region

#Region "Other event handlers"
'show what came back from last PIC in chain:
'(useful for diagnosing comm problems or performance tuning)
	Private Sub selectedPort_DataReceived(ByVal sender As Object, ByVal e As SerialDataReceivedEventArgs) Handles m_selectedPort.DataReceived
'TODO: auto-check firmware version and PIC status?
		Dim port As SerialPort = DirectCast(sender, SerialPort)
		Dim rcvlen As Integer = port.BytesToRead, inlen As Integer = rcvlen
		If (inlen < 0) Or (inlen > Me.m_inbuf.Length) Then inlen = Me.m_inbuf.Length 'paranoid; limit size in case something went wrong
		port.Read(Me.m_inbuf, 0, inlen)
		Me.m_rcvbuf = DumpHex(Me.m_inbuf, inlen, 80) 'save it to see later
		If Me.m_trace Then LogMsg(String.Format("Chipiplex[rcv] got {0} bytes: ", rcvlen) & Me.m_rcvbuf)
	End Sub
#End Region

#Region "Implementation/helpers"
	Private Property State As RunState '2.5.0.8
		Get
			Return Me.m_state
		End Get
		Set(ByVal value As RunState)
			Me.m_state = value
			If (value = RunState.Stopping) Then
				Me.m_eventTrigger.Set
				Me.m_eventTrigger.Close
				Me.m_eventTrigger = Nothing
			End If
		End Set
	End Property

	Private Sub EventThread()
		Me.State = RunState.Running '2.5.0.8
		Me.m_eventTrigger = New AutoResetEvent(False) '2.5.0.8
		Try 
'			Do While Me.m_running '2.1.1.0
			Do While (Me.State = RunState.Running) '2.5.0.8
'				Me.m_threadPosition = 0 '2.1.1.0
				Me.m_eventTrigger.WaitOne
'				Me.m_threadPosition = 1
'				Me.m_threadPosition = 2
				Try '2.5.0.8
					Me.FireEvent
				Catch exception1 As TimeoutException '2.5.0.8
				End Try '2.5.0.8
'				Me.m_threadPosition = 3
			Loop
'			Me.m_threadPosition = 4
		Catch exception As Exception
'			Me.m_threadPosition = 4
			If (Me.State = RunState.Running) Then Me.State = RunState.Stopping '2.5.0.8
		Finally '2.5.0.8
			Me.State = RunState.Stopped '2.5.0.8
		End Try
	End Sub

	Private Sub FireEvent()
'		If Not Me.m_selectedPort.IsOpen Then Me.m_selectedPort.Open
		If (Me.m_protocolVersion = 1) Then
			Me.Protocol1Event(Me.m_channelValues)
		ElseIf (Me.m_protocolVersion = 2) Then
			Me.Protocol2Event(Me.m_channelValues)
#If WANT_CHIPIPLEXING
		ElseIf (Me.m_protocolVersion = 3) Then
			Me.ChipiplexedEvent(Me.m_channelValues)
#End If
		Else
			ReportError(String.Format("Unhandled protocol: {0}.  Please report this to the author(s).", Me.m_protocolVersion))
		End If
	End Sub

'no need to pass by val since we already have a local copy: -dj
'	Private Sub Protocol1Event(ByVal channelValues As Byte())
	Private Sub Protocol1Event(ByRef channelValues As Byte())
		Dim length As Integer = channelValues.Length
		Dim count As Integer = 2
		Dim num3 As Integer = ((2 + (2 * length)) + ((2 + (2 * length)) / 100)) 'double length + 1 extra pad byte for each 100 bytes
		If (Me.m_p1Packet.Length < num3) Then Me.m_p1Packet = New Byte(num3  - 1) {}
		Me.m_p1Packet(0) = Renard_SyncByte '&h7E
		Me.m_p1Packet(1) = Renard_CmdAdrsByte '&h80
'		Me.m_threadPosition = 10
		Dim i As Integer
		For i = 0 To length - 1
'replace special protocol bytes with alternate bytes to avoid protocol errors:
'a slight change to dimmer value is not noticeable
			If (m_channelValues(i) = Renard_PadByte) Then '&h7D
'				Me.m_threadPosition = 11
				Me.m_p1Packet(count) = Renard_SafeByte '&h7C
			ElseIf (channelValues(i) = Renard_SyncByte) Then '&h7E
'				Me.m_threadPosition = 12
				Me.m_p1Packet(count) = Renard_SafeByte '&h7C
			ElseIf (channelValues(i) = Renard_EscByte) Then '&h7F
'				Me.m_threadPosition = 13
				Me.m_p1Packet(count) = Renard_CmdAdrsByte '&h80
			Else
'				Me.m_threadPosition = 14
				Me.m_p1Packet(count) = channelValues(i)
			End If
			count += 1
			If ((count Mod 100) = 0) Then
'				Me.m_threadPosition = 15
				Me.m_p1Packet(count) = Renard_PadByte '&h7D
				count += 1
			End If
'			Me.m_threadPosition = 16
		Next i
'		Me.m_threadPosition = 17
'		If Me.m_running Then 2.1.1.0
			Do While ((Me.m_selectedPort.WriteBufferSize - Me.m_selectedPort.BytesToWrite) <= count)
				Thread.Sleep(5) '(10) '2.5.0.8
			Loop
'			Me.m_threadPosition = 18
			Me.m_selectedPort.Write(Me.m_p1Packet, 0, count)
'			Me.m_threadPosition = 19
'        End If
'        Me.m_threadPosition = 20
		If Me.m_trace Then LogMsg(String.Format("Protocol1 write {0} -> {1} bytes : ", length, count) & DumpHex(Me.m_p1Packet, count, 20))
	End Sub

'not sure if this one is still used: -dj
'no need to pass by val since we alredy have a local copy -dj
'	Private Sub Protocol2Event(ByVal channelValues As Byte())
	Private Sub Protocol2Event(ByRef channelValues As Byte())
		Dim num3 As Byte = Renard_CmdAdrsByte '&h80
		Dim length As Integer = channelValues.Length
		Dim array As Byte() = New Byte(8  - 1) {}
		Dim i As Integer = 0
		Do While (i < length)
			Dim num2 As Integer = Math.Min(CInt((i + 7)), CInt((length - 1)))
			num3 = CByte((num3 + 1))
			Me.m_p2Packet(1) = num3
			If (num2 >= (length - 1)) Then Me.m_p2Zeroes.CopyTo(Me.m_p2Packet, 3)
			System.Array.Clear(array, 0, 8)
			Dim index As Integer = i
			Do While (index <= num2)
				Dim num8 As Byte = channelValues(index)
				Dim num9 As Byte = -num8
				If ((num8 >= 1) AndAlso (num8 <= 8)) Then
					array((num8 - 1)) = 1
				ElseIf ((num9 >= 1) AndAlso (num9 <= 8)) Then
					array((num9 - 1)) = 1
				End If
				index += 1
			Loop
			Dim num7 As Byte = CByte((1 + System.Array.IndexOf(Of Byte)(array, 0)))
			Me.m_p2Packet(2) = num7
			index = i
			Dim num5 As Integer = 3
			Do While (index <= num2)
				Me.m_p2Packet(num5) = CByte((channelValues(index) - num7))
				index += 1
				num5 += 1
			Loop
'			If Me.m_running Then '2.1.1.0
			Me.m_selectedPort.Write(Me.m_p2Packet, 0, num5)
			If Me.m_trace Then LogMsg(String.Format("Protocol2 write {0} bytes from ofs {1} of {2}: ", num5, i, length) & DumpHex(Me.m_p2Packet, num5, 20))
			i = (i + 8)
		Loop
	End Sub

#If WANT_CHIPIPLEXING Then
'added new code for chipiplexing: -dj
'data packet format for each PIC (group of 56 channels) is:
'  sync byte (same as Renard)
'  command byte (same as Renard)
'  config byte: pulse width, PWM, debug, etc
'  then repeating groups of:
'      brightness level
'      row index
'      columns, columns, ... (one byte for each row)
'  0 for end-of-list indicator
	Private Sub ChipiplexedEvent(ByRef channelValues As Byte())
		on error goto errh
		If channelValues.Length <> Me.m_numch Then Throw New Exception(String.Format("event received {0} channels but configured for {1}", channelValues.Length, Me.m_numch))
		Me.total_inlen += channelValues.Length
		Me.total_evts += 1
'TODO: show more or less data
		If Me.m_trace Then LogMsg(String.Format("Chipiplex[{0}] {1} bytes in: ", Me.total_evts, channelValues.Length) & DumpHex(channelValues, channelValues.Length, 80))
'format output buf for each PIC (groups of up to 56 chipiplexed channels):
		Dim buflen As Integer, protolen As Integer 
		Dim overflow_retry As Integer, overflow_adjust As Integer, merge_retry As Integer 
		Dim brlevel As Integer, brinx As Integer, row As Integer, col As Integer 
		Dim i As Integer
		Dim pic As Integer
'NOTE to style critics: I hate line numbers but I also hate cluttering up source code with a lot of state/location info
'I'm only using line#s because they are the most compact way to show the location of errors (kinda like C++'s __LINE__ macro), and there is no additional overhead
12:
'send out data to PICs in *reverse* order (farthest PICs first, closest PICs last):
'this will allow downstream PICs to get their data earlier and start processing while closer PICs are getting their data
'if we send data to closer PICs first, they will have processed all their channels before downstream PICs even get theirs, which could lead to partially refreshed frames
'at 115k baud, 100 chars take ~ 8.7 mec, which is > 1 AC half-cycle at 60 Hz; seems like this leads to slight sync problems
'TODO: make forward vs. reverse processing a config option
		For pic = Me.m_numch - (Me.m_numch - 1) Mod 56 To 1 Step -56 'TODO: config #channels per PIC?
			Dim status As String = "" 'show special processing flags in trace file
			Dim numbr As Integer = 0 '#brightness levels
			Dim numrows As Integer = 0, numcols As Integer '#rows, columns occupied for this frame
			For i = 0 To 255 'initialize brightness level index to all empty
				Me.m_brindex(i) = 0 'empty
			Next i
14:
			Dim ch As Integer
			For ch = pic To pic+55 'each channel for this PIC (8*7 = 56 for a regular Renard PIC); 1-based
				If ch > Me.m_numch Then Exit For
				brlevel = channelValues(ch-1)
16:
				If brlevel < Me.m_minbright Then 'don't need to send this channel to controller (it's off)
					If brlevel = 0 then
						Me.total_realnulls += 1
					Else
						Me.total_nearnulls += 1
					End If
				Else 'need to send this channel to the controller (it's not off)
18:
					Dim overflow_limit As Integer = Me.m_closeness 'allow overflow into this range of values
					If brlevel >= Me.m_maxbright Then 'treat as full on
						If brlevel < 255 Then Me.total_nearfulls += 1
						overflow_limit += Me.m_fullbright - Me.m_maxbright 'at full-on, expand the allowable overflow range
						brlevel = Me.m_fullbright '255
					End If
					row = (ch - pic)\7 'row# relative to this PIC (0..7); do not round up
					col = (ch - pic) Mod 7 'column# relative to this row
					If col >= row Then col += 1 'skip over row address line (chipiplexing matrix excludes diagonal row/column down the middle)
#If BOARD1_INCOMPAT Then 'whoops; wired the pins differently between boards; rotate them here rather than messing with channel reordering in Vixen
'					Dim svch As Integer = ch - pic: ch = 0
'					If svch And &h80 Then ch += &h10
'					If svch And &h40 Then ch += 1
'					If svch And &h20 Then ch += &h40
'					If svch And &h10 Then ch += &h20
'					If svch And 8 Then ch += &h80
'					If svch And 4 Then ch += 8
'					If svch And 2 Then ch += 4
'					If svch And 1 Then ch += 2
'					ch += pic
					Select Case row
						Case 0: row = 4'3
						Case 1: row = 2'7
						Case 2: row = 3'1
						Case 3: row = 0'2
						Case 4: row = 5'0
						Case 5: row = 6'4
						Case 6: row = 7'5
						Case 7: row = 1'6
					End Select
					Select Case col
						Case 0: col = 4'3
						Case 1: col = 2'7
						Case 2: col = 3'1
						Case 3: col = 0'2
						Case 4: col = 5'0
						Case 5: col = 6'4
						Case 6: col = 7'5
						Case 7: col = 1'6
					End Select
#End If
20:
					For overflow_retry = 0 To 2 * overflow_limit
						If (overflow_retry And 1) = 0 Then 'try next higher (brighter) level
							overflow_adjust = brlevel + overflow_retry\2 'do not round up
							If overflow_adjust > Me.m_fullbright Then Continue For
						Else 'try next lower (dimmer) value
							overflow_adjust = brlevel - overflow_retry\2 'do not round up
							If overflow_adjust < Me.m_minbright Then Continue For
						End If
22:
						brinx = Me.m_brindex(overflow_adjust)
						If brinx = 0 Then 'allocate a new brightness level
							If Me.IsRenardByte(overflow_adjust) Then Me.num_avoids += 1: status &= "@": Continue For 'avoid special protocol chars
							numbr += 1: brinx = numbr 'entry 0 is used as an empty; skip it
							Me.m_brindex(overflow_adjust) = brinx 'indexing by brightness level automatically sorts them by brightness
'							Me.m_levels(brinx).level = brlevel 'set brightness level for this entry
							Me.m_rowindex(brinx) = 0 'no rows yet for this brightness level
'							Me.m_levels(brinx).numrows = 0
							For i = 0 To 7 'no columns yet for this brightness level
								Me.m_columns(brinx, i) = 0
							Next i
						End If
24:
						If Me.m_columns(brinx, row) = 0 Then 'allocate a new row
'							If Me.m_levels(brinx).numrows >= Me.m_maxrpl Then Continue For 'this level is full; try another one that is close (lossy dimming)
'enforce maxRPL here to avoid over-fillings brightness levels; lossy dimming will degrade with multiple full levels near each other
							If Me.NumBits(Me.m_rowindex(brinx)) >= Me.m_maxrpl Then Continue For 'this level is full; try another one that is close (lossy dimming)
'maxRPL can't be more than 2 in this version, so we don't need to check for reserved bytes here (the only one that matters has 7 bits on)
'							If Me.IsRenardByte(overflow_adjust) Then Me.num_avoids += 1: status &= "@": Continue For 'avoid special protocol chars
							Me.m_rowindex(brinx) = Me.m_rowindex(brinx) Or 1<<(7 - row)
'							Me.m_levels(brinx).numrows += 1
							numrows += 1
						End If
						Dim newcols As Byte = Me.m_columns(brinx, row) Or 1<<(7 - col)
						If Me.IsRenardByte(newcols) Then Me.num_avoids += 1: status &= "@": Continue For 'avoid special protocol chars; treat as full row
						Me.m_columns(brinx, row) = newcols 'Me.m_columns(brinx, row) Or 1<<(7 - col)
						If overflow_retry <> 0 Then Me.total_rowsoverflowed += 1: status &= "*"
						Exit For
					Next overflow_retry
26:
					If overflow_retry > 2 * overflow_limit Then
						If Me.m_nodiscard Then Throw New Exception(String.Format("Unable to overflow full row after {0} tries.", 2 * overflow_limit))
						Me.total_rowsdropped += 1
						status &= "-"
					End If
				End If
			Next ch
28:
#If False Then 'obsolete merge/reduction code
'			Dim brlimit As Integer = Me.m_maxbright + 8\Me.m_maxrpl - 1 'allow for overflow of max bright level
'			If brlimit > 255 Then brlimit = 255 'don't overshoot real top end of brightness range
			For merge_retry = 0 To Me.m_closeness
				If 2+2 + 2*numbr + numrows <= Me.m_outbuf.Length Then Exit For 'no need to merge brightness levels
				If merge_retry = 0 Then Continue For 'dummy loop entry to perform length check first time
				For brlevel = Me.m_minbright To Me.m_fullbright 
					Dim tobrinx As Byte = Me.m_brindex(brlevel)
					If tobrinx = 0 Then Continue For
					Dim otherlevel As Integer 
					Dim mergelimit As Integer = brlevel + merge_retry
					if brlevel >= Me.m_maxbright Then mergelimit = 255
30:
					For otherlevel = brlevel - merge_retry To merge_limit Step 2*merge_retry
						If (otherlevel < Me.m_minbright) Or (otherlevel > brlimit) Then Continue For
						Dim frombrinx As Byte = Me.m_brindex(otherlevel)
						If (frombrinx = 0) or (Me.numbits(Me.m_rowindex(tobrinx) And Me.m_rowindex(frombrinx)) = 0) Then continue for 'can't save some space by coalescing
						numrows -= Me.numbits(Me.m_rowindex(tobrinx) And Me.m_rowindex(frombrinx))
						Me.m_rowindex(tobrinx) = Me.m_rowindex(tobrinx) Or Me.m_rowindex(frombrinx) 'merge row index
						For row = 0 To 7 'merge columns
							Me.m_columns(tobrinx, row) = Me.m_columns(tobrinx, row) Or Me.m_columns(frombrinx, row)
						Next row
32:
						Me.m_brindex(otherlevel) = 0 'hide entry so it won't be used again
						numbr -= 1
						Me.total_levelsmerged += 1
						status &= "^"
						If 2+2 + 2*numbr + numrows <= Me.m_outbuf.Length Then Exit For 'no need to coalesce more brightness levels
						Continue For
					Next otherlevel
				Next brlevel
			Next merge_retry
34:
			If merge_retry > Me.m_closeness Then
				If Me.m_nodiscard Then Throw New Exception(String.Format("Unable to merge brightness level after {0} tries.", 2*Me.m_closeness))
				Me.total_levelsdropped += 1
				status &= "X"
			End If
#End If
			Dim prevlevel As Integer = 0, prevbrinx As Integer 
'TODO: use MaxData length here
			While 3+2 + 2*numbr + numrows > Me.m_outbuf.Length 'need to merge brightness levels
'this is expensive, so config params should be set to try to avoid it
'TODO: this could be smarter; prioritize and then merge rows that are closest or sparsest first
				For brlevel = Me.m_minbright To Me.m_fullbright 
					brinx = Me.m_brindex(brlevel)
					If brinx = 0 Then Continue For
					If prevlevel < 256 Then 'merge row with previous to save space
30:
						If prevlevel = 0 Then prevlevel = brlevel: prevbrinx = brinx: Continue For 'need 2 levels to compare
						If (prevlevel + Me.m_closeness < brlevel) And (prevlevel + Me.m_closeness < Me.m_maxbright) Then Continue For 'too far apart to merge
						Dim newrows As Byte = Me.m_rowindex(brinx) Or Me.m_rowindex(prevbrinx) 'merge row index
						If Me.NumBits(newrows) > Me.m_maxrpl Then
'							If Not String.IsNullOrEmpty(Me.m_logfile) Then LogMsg("row " & brlevel & "->" & brinx & " full; can't merge")
							Continue For 'can't merge; row is already full
						End If
						If Me.IsRenardByte(newrows) Then Me.num_avoids += 1: status &= "@": Continue For 'can't merge (would general special protocol chars)
						Dim newcols(8 - 1) As Byte
						For row = 0 To 7 'merge columns
							newcols(row) = Me.m_columns(brinx, row) Or Me.m_columns(prevbrinx, row)
							If Me.IsRenardByte(newcols(row)) Then Me.num_avoids += 1: status &= "@": Exit For 'can't merge (would general special protocol chars)
						Next row
						If row <= 7 Then Continue For 'can't merge (would general special protocol chars)
						numrows -= Me.NumBits(Me.m_rowindex(brinx) And Me.m_rowindex(prevbrinx))
						Me.m_rowindex(brinx) = newrows 'Me.m_rowindex(brinx) Or Me.m_rowindex(prevbrinx) 'merge row index
						For row = 0 To 7 'merge columns
							Me.m_columns(brinx, row) = newcols(row) 'Me.m_columns(brinx, row) Or Me.m_columns(prevbrinx, row)
						Next row
						If Me.m_trace Then LogMsg(String.Format("merged level {0} with {1}", prevlevel, brlevel))
						Me.m_brindex(prevlevel) = 0 'hide this entry so it won't be used again
						Me.total_levelsmerged += 1
						status &= "^"
					Else 'last try; just drop the row
						If Me.m_trace Then LogMsg(String.Format("dropped level {0}", brlevel))
						Me.m_brindex(brlevel) = 0 'hide this entry so it won't be used again
						numrows -= Me.NumBits(Me.m_rowindex(brinx))
						Me.total_levelsdropped += 1
						status &= "!"
					End If
32:
					numbr -= 1
					If 3+2 + 2*numbr + numrows <= Me.m_outbuf.Length Then Exit While 'no need to merge/drop more brightness levels
				Next brlevel
				If Me.m_nodiscard Then Throw New Exception(String.Format("Unable to merge brightness level after {0} tries.", 2*Me.m_closeness))
				prevlevel = 256 'couldn't find rows to merge, so start dropping them
34:
			End While
'TODO: resend data after a while even if it hasn't changed?
'no, unless dup:			If numbr < 1 Then Continue For 'no channels to send this PIC
'format data to send to PIC:
'TODO: send more than 1 SYNC the first time, in case baud rate not yet selected
			buflen = 0
			protolen = 0
'			If pic = 1 Then 'only needed for first PIC in chain?  TODO
				Me.m_outbuf(buflen) = Renard_SyncByte: buflen += 1 '&h7E
				protolen += 1
'			End If
			Me.m_outbuf(buflen) = Renard_CmdAdrsByte + pic\56: buflen += 1 '&h80
'			Me.m_outbuf(buflen) = Renard_PadByte: buflen += 1 '&h7D 'TODO?
			Me.m_outbuf(buflen) = Me.m_cfg: buflen += 1 'config byte comes first
			protolen += 2
			For brlevel = Me.m_fullbright To Me.m_minbright Step -1 'send out brightess levels in reverse order (brightest first)
36:
				brinx = Me.m_brindex(brlevel)
				If brinx = 0 Then Continue For
38:
				If (brlevel = 0) Or (Me.m_rowindex(brinx) = 0) Then SentBadByte(IIf(brlevel = 0, brlevel, Me.m_rowindex(brinx)), buflen, pic\56) 'paranoid
				If Me.IsRenardByte(brlevel) Then SentBadByte(brlevel, buflen, pic\56) 'paranoid
				If buflen < m_outbuf.Length Then Me.m_outbuf(buflen) = brlevel 'send brightness level; avoid outbuf overflow
				buflen += 1 'maintain count, even if outbuf overflows (so we know how bad is was)
				If Me.IsRenardByte(Me.m_rowindex(brinx)) Then SentBadByte(Me.m_rowindex(brinx), buflen, pic\56) 'paranoid
				If buflen < m_outbuf.Length Then Me.m_outbuf(buflen) = Me.m_rowindex(brinx) 'send row index byte; avoid outbuf overflow
				buflen += 1 'maintain count, even if outbuf overflows (so we know how bad is was)
40:
				For row = 0 To 7
'					If buflen >= m_outbuf.Length Then LogMsg("ERROR: outbuf overflow at row " & row & " of brinx " & brinx & ", brlevel " & brlevel & ", thought I needed 4+2*" & numbr & "+" & numrows & "=" & (2+2 + 2*numbr + numrows) & ", outbuf so far is " & buflen & ":" & DumpHex(Me.m_outbuf, buflen, 80))
'					If Not String.IsNullOrEmpty(Me.m_logfile) Then LogMsg("err 9 debug: brinx " & brinx & ", row " & row & ", buflen " & buflen)
					If Me.m_columns(brinx, row) <> 0 Then 'send columns for this row
						If Me.IsRenardByte(Me.m_columns(brinx, row)) Then SentBadByte(Me.m_columns(brinx, row), buflen, pic\56) 'paranoid
						If buflen < m_outbuf.Length Then Me.m_outbuf(buflen) = Me.m_columns(brinx, row) 'avoid outbuf overflow
						buflen += 1 'maintain count, even if outbuf overflows (so we know how bad is was)
					End If
					numcols += Me.NumBits(Me.m_columns(brinx, row))
				Next row
			Next brlevel
42:
			If buflen < m_outbuf.Length Then Me.m_outbuf(buflen) = 0: 'end of list indicator; avoid outbuf overflow
			buflen += 1 'maintain count, even if outbuf overflows (so we know how bad is was)
'			If pic = 1 Then 'last packet; send sync to kick out of packet receive loop (to allow console/debug commands); send this on all, in case last one is dropped as a dup
'TODO: trailing Sync might not be needed any more
				If buflen < m_outbuf.Length Then Me.m_outbuf(buflen) = Renard_SyncByte: '&h7E; avoid outbuf overflow
				buflen += 1 'maintain count, even if outbuf overflows (so we know how bad is was)
				protolen += 1
'			End If
			If buflen > m_outbuf.Length Then 'outbuf overflowed; this is a bug if it happens
				Throw New Exception(String.Format("ERROR: outbuf overflow, thought I needed 5+2*{0}+{1}={2}, but really needed {3}:", numbr, numrows, 3+2 + 2*numbr + numrows, buflen) & DumpHex(Me.m_outbuf, Math.Min(buflen, Me.m_outbuf.Length), 80))
			End If
			If (pic\56 = Me.m_bufdedup) And (buflen = Me.m_prevbuflen) Then 'compare current buffer to previous buffer
				For i = 1 To buflen
					If Me.m_outbuf(i-1) <> Me.m_prevbuf(i-1) Then Exit For 'need to send output buffer
				Next i
				If i > buflen Then 'outbuf was same as last time; skip it
					Me.num_dups += 1
					If Me.m_trace Then LogMsg(String.Format("Chipiplex[{0}]= duplicate buffer on {1} discarded", Me.total_evts, Me.m_bufdedup))
					Continue For
				End If
			End If
			Me.total_outlen += buflen
			Me.protocol_outlen += protolen
			Me.total_levels += numbr
			Me.total_rows += numrows
			Me.total_columns += numcols
			If (numbr > 0) And (numbr < Me.min_levels) Then Me.min_levels = numbr
			If numbr > Me.max_levels Then Me.max_levels = numbr
			If (numrows > 0) And (numrows < Me.min_rows) Then Me.min_rows = numrows
			If numrows > Me.max_rows Then Me.max_rows = numrows
			If (numcols > 0) And (numcols < Me.min_columns) Then Me.min_columns = numcols
			If numcols > Me.max_columns Then Me.max_columns = numcols
			If (buflen > 0) And (buflen < Me.min_outlen) Then Me.min_outlen = buflen
			If buflen > Me.max_outlen Then Me.max_outlen = buflen
44:
'TODO: append multiple bufs together, and only write once
			Me.m_selectedPort.Write(Me.m_outbuf, 0, buflen)
			If pic\56 = Me.m_bufdedup Then 'save current buffer to compare to next time
				For i = 1 To buflen 
					Me.m_prevbuf(i-1) = Me.m_outbuf(i-1)
				Next i
				Me.m_prevbuflen = buflen
			End If
46:
			If Me.m_trace Then LogMsg(String.Format("Chipiplex[{0}]{1} {2}+{3} bytes out: ", Me.total_evts, status, protolen, buflen - protolen) & DumpHex(Me.m_outbuf, buflen, 80))
		Next pic
		Exit Sub
	errh:
		ReportError("ChipiplexedEvent")
	End Sub
#End If
	
'write message to log file with relative timestamp:
	Private Sub LogMsg(ByRef buf As String)
		On Error Resume Next 'just ignore errors if they occur, rather than getting into an infinite loop
		If String.IsNullOrEmpty(Me.m_logfile) Then Exit Sub
		Dim tdiff As TimeSpan = DateTime.Now.Subtract(Me.m_starttime)
		Dim msg as String = String.Format("[T+{0:D}:{1:D2}:{2:D2}.{3:D3}] " & buf & ChrW(13) & ChrW(10), New Object() { tdiff.Hours, tdiff.Minutes, tdiff.Seconds, tdiff.Milliseconds})
'		Dim outstream As filestream = File.open(Me.m_logfile, FileMode.OpenOrCreate, FileAccess.Write, FileShare.ReadWrite)
'		Dim buf As Byte() = system.Text.Encoding.Default.GetBytes(msg)
'		If outstream.length > 0 Then stream.position = stream.length - 1
'		outstream.write(buf, 0, msg.length - 1)
'		outstream.close
		Monitor.Enter(LogFileContention)
		System.IO.File.AppendAllText(Me.m_logfile, msg) 'AppendAllText is NOT THREAD-SAFE, so access must be managed
		Monitor.Exit(LogFileContention)
	End Sub
	
'throw error if invalid byte sent:
	Private Sub SentBadByte(ByVal badbyte as Byte, ByVal ofs As Integer, ByVal pic As Integer)
		Throw New Exception(String.Format("Tried to send empty or reserved byte {0} at ofs {1} for pic {2}. Please report this.", badbyte, ofs, pic))
	End Sub

'avoid Renard special byte values, so chipiplexing + non-chipiplexing can be intermixed on same comm line:
	Private Function IsRenardByte(ByVal val As Byte) As Boolean
'		If (val >= Me.RenardReserved_lowest) And (val <= Me.RenardReserved_highest) Then Return True
		If val = Renard_SyncByte Then Return True
		If Me.m_renbytes And ((val = Renard_SafeByte) Or (val = Renard_PadByte) Or (val = Renard_EscByte)) Then Return True
		Return False
	End Function

'write value to sequence, read back and verify:
	Private Sub ParanoidSetInteger(ByVal name As String, ByVal value As Integer)
		Dim verify As Integer 
		Me.m_setupData.SetInteger(Me.m_setupNode, name, value)
		verify = Me.m_setupData.GetInteger(Me.m_setupNode, name, value - 1)
		If verify <> value Then Throw New Exception(String.Format("setupData.SetInteger failed: got {0}, expected {1}", verify, value)) 'paranoid: verify file I/O
	End Sub

'display error message and write to log:
	Private Sub ReportError(ByVal where As String)
		If Erl <> 0 Then where &= " at " & Erl 'show line# if present
		Dim msg As String = Err.Description & String.Format(" (error {0} in {1})", Err.Number, where)
		LogMsg("ERROR: " & msg)
		If Not Me.m_errpause Then Return
'		If (Me.State = RunState.Running) Then
'			Me.State = RunState.Stopping
'			Do While (Me.State <> RunState.Stopped)
'				Thread.Sleep(5)
'			Loop
'		End If
'		If Not Me.m_selectedPort Is Nothing AndAlso Me.m_selectedPort.IsOpen Then Me.m_selectedPort.Close
'this isn't quite right yet, but at least it stops :)
		If MsgBox(msg & vbCrLf & "Continue execution?", MsgBoxStyle.Exclamation + MsgBoxStyle.ApplicationModal + MsgBoxStyle.OkCancel, "ERROR") <> MsgBoxResult.Ok Then Thread.CurrentThread.Abort 
	End Sub

'dump a byte array as hex values:
	Private Function DumpHex(ByRef buf As Byte(), ByVal ActualLen As Integer, ByVal MaxLen As Integer) As String
		Dim hexbuf As StringBuilder = New StringBuilder 
		Dim trunc As Boolean = False
		While ActualLen > 1
			If buf(ActualLen - 1) <> 0 Then Exit While
			trunc = True
			ActualLen -= 1
		End While
		Dim i As Integer
		For i = 1 To ActualLen
			If i > MaxLen Then trunc = True: Exit For
			hexbuf.AppendFormat("{0:X2} ", buf(i-1))
		Next i
		If trunc Then hexbuf.Append(" ...")
		Return hexbuf.ToString
	End Function
	
'avoid divide-by-0 errors:
	Private Function SafeDiv(ByVal num As Integer, ByVal den As Integer) As Integer
		If den <> 0 Then
			Return num/den 'rounded divide; use num\den to avoid rounding
		Else
			Return 0 'avoid divide by 0 error if no stats collected yet
		End If
	End Function
#End Region
End Class
